%function [cX, cY, cZ] = Reconstruction3D(filename_cal, filename_corr, output_file)
%
%  given a calibration file and a file with a list of correspondences
%  reconstructs the points, store it in prn format in the indicated
%  output file and fits a 3D plane in the cloud of points and another
%  to the normalized points
%
%
function [cX, cY, cZ] = Reconstruction3D(filename_cal, filename_corr, output_file, min_depth, max_depth)

[Tc_i, Tc_e, Tp_i, Tp_e] = LoadCalibration(filename_cal);
[xc, yc, xp] = LoadCorrespondences(filename_corr);


num_corr = length(xc);

wX = [];
wY = [];
wZ = [];

cX = [];
cY = [];
cZ = [];

Ac = Tc_i*Tc_e;
Ap = Tp_i*Tp_e;

P=zeros(3,3);
F=zeros(3,1);

for i=1:num_corr,
    P(1,:) = [Ap(1,1)-Ap(3,1)*xp(i)    Ap(1,2)-Ap(3,2)*xp(i)     Ap(1,3)-Ap(3,3)*xp(i)];
    P(2,:) = [Ac(1,1)-Ac(3,1)*xc(i)    Ac(1,2)-Ac(3,2)*xc(i)     Ac(1,3)-Ac(3,3)*xc(i)];  
    P(3,:) = [Ac(2,1)-Ac(3,1)*yc(i)    Ac(2,2)-Ac(3,2)*yc(i)     Ac(2,3)-Ac(3,3)*yc(i)];
    
    F(1,:) = [Ap(3,4)*xp(i) - Ap(1,4)];
    F(2,:) = [Ac(3,4)*xc(i) - Ac(1,4)];
    F(3,:) = [Ac(3,4)*yc(i) - Ac(2,4)];
    
    wPoint = inv(P)*F;  
    
    wX = [wX; wPoint(1)];
    wY = [wY; wPoint(2)];
    wZ = [wZ; wPoint(3)];
    
    %we express the points in the camera frame:
    cPoint = Tc_e*[wX(i); wY(i); wZ(i); 1];
    
    if (cPoint(3)>min_depth)&(cPoint(3)<max_depth) %&(cPoint(1)<130)
        cX =[cX; cPoint(1)];   
        cY =[cY; -cPoint(2)];  
        cZ =[cZ; cPoint(3)];   
    end
end

%store the reconstruction in output.prn:
output = fopen(output_file,'w+');
%save header:
fprintf(output, '%d\n', length(cX));
%save points: x y z
for i=1:length(cX),
    fprintf(output, '0 0 0%12.5f %12.5f %12.5f\n', cX(i), cY(i), cZ(i));
end
fclose(output);

figure;
hold on;
plot3(cX, cY, cZ, '.');
axis equal;

disp('====================================================');
num_points = length(cX)

%fit a 3D plane:
[centroid, normal] = lsplane([cX cY cZ]);
normal = normal/norm(normal);

A = normal(1);
B = normal(2);
C = normal(3);
D = centroid'*normal;

distances = [];
for i=1:length(cX),
    distances = [distances; abs(A*cX(i)+B*cY(i)+C*cZ(i)-D)/sqrt(A^2+B^2+C^2)];
end

distances = sort(distances);
distances = distances(1:length(distances)-1);

mean_distance = mean(distances);
median_distance = median(distances);
max_distance = max(distances);
std_distance = std(distances);

disp('====================================================');
disp('Non-normalized XYZ points:');
disp('--------------------------');
disp('Plane normal:');
n = [A B C]
disp('Distance to the origin (D):');
D
disp('mean distance to the fitted plane (mm):');
mean_distance
disp('median distance to the fitted plane (mm):');
median_distance
disp('max distance to the fitted plane (mm):');
max_distance
disp('standard deviation of the distance to the fitted plane (mm):');
std_distance

% %-------------------------------------
% %  POINTS NORMALIZATION:
% %
% %  translate the centroid of the 
% %  cloud of points to the origin and
% %  divide each point by the maximum
% %  norm.
% %-------------------------------------
% cX = cX - centroid(1);
% cY = cY - centroid(2);
% cZ = cZ - centroid(3);
% max_norm = max(norm([cX cY cZ]));
% cX = cX / max_norm;
% cY = cY / max_norm;
% cZ = cZ / max_norm;
% figure;
% hold on;
% plot3(cX, cY, cZ, '.');
% axis equal;
% 
% %re-fit a 3D plane:
% [centroid, normal] = lsplane([cX cY cZ]);    
% normal = normal/norm(normal);
% 
% if (normal(3)<0)
%     normal = -normal;
% end
% A = normal(1);
% B = normal(2);
% C = normal(3);
% D = 0;
% 
% distances = [];
% for i=1:length(cX),
%     distances = [distances; abs(A*cX(i)+B*cY(i)+C*cZ(i)-D)/sqrt(A^2+B^2+C^2)];
% end
% mean_distance = mean(distances);
% std_distance = std(distances);
% disp('====================================================');
% disp('Normalized XYZ points:');
% disp('--------------------------');
% disp('Plane normal:');
% n = [A B C]
% disp('Distance to the origin (D):');
% D
% disp('mean distance to the fitted plane:');
% mean_distance
% disp('standard deviation of the distance to the fitted plane:');
% std_distance
% disp('====================================================');
